from geolysis.foundation import Foundation
from geolysis.utils import round_

from ._core import AllowableBearingCapacity


class BowlesABC4PadFoundation(AllowableBearingCapacity):
    r"""Allowable bearing capacity for pad foundation on cohesionless
    soils according to `Bowles (1997)`.

    See [implementation](../formulas/allowable-bearing-capacity.md/#bowles-bearing-capacity-for-pad-foundation)
    for more details on bearing capacity equation used.

    """

    def __init__(
        self,
        corrected_spt_n_value: float,
        tol_settlement: float,
        foundation_size: Foundation,
    ) -> None:
        """
        :param corrected_spt_n_value: Statistical average of corrected
                                      SPT N-value (55% energy with
                                      overburden pressure correction)
                                      within the foundation influence
                                      zone i.e `0.5B` to `2B`.
        :param tol_settlement: Tolerable settlement of foundation (mm).
        :param foundation_size: Size of the foundation.
        """
        super().__init__(
            corrected_spt_n_value=corrected_spt_n_value,
            tol_settlement=tol_settlement,
            foundation_size=foundation_size,
        )

    @round_(ndigits=2)
    def _bearing_capacity(self) -> float:
        """
        Calculate the allowable bearing capacity of the pad foundation.
        """
        n_corr = self.corrected_spt_n_value
        width = self.foundation_size.width

        if width <= 1.2:
            return 19.16 * n_corr * self._fd() * self._sr()

        return (
            11.98
            * n_corr
            * ((3.28 * width + 1) / (3.28 * width)) ** 2
            * self._fd()
            * self._sr()
        )


class BowlesABC4MatFoundation(BowlesABC4PadFoundation):
    r"""Allowable bearing capacity for mat foundation on cohesionless
    soils according to `Bowles (1997)`.

    See [implementation](../formulas/allowable-bearing-capacity.md/#bowles-bearing-capacity-for-mat-foundation)
    for more details on bearing capacity equation used.

    """

    @round_(ndigits=2)
    def _bearing_capacity(self) -> float:
        """
        Calculate the allowable bearing capacity of the mat foundation.
        """
        n_corr = self.corrected_spt_n_value
        return 11.98 * n_corr * self._fd() * self._sr()



import enum
from abc import ABC, abstractmethod
from typing import Annotated, Final, Optional, Sequence

from func_validator import (
    DependsOn,
    MustBeBetween,
    MustBeGreaterThanOrEqual,
    MustBeMemberOf,
    MustBeNonNegative,
    MustBePositive,
    MustHaveLengthGreaterThan,
    MustHaveValuesBetween,
    validate_params,
)

from .foundation import Foundation
from .utils import AbstractStrEnum, isclose, isinf, log10, mean, round_, sqrt

__all__ = [
    "SPT",
    "HammerType",
    "SamplerType",
    "EnergyCorrection",
    "GibbsHoltzOPC",
    "BazaraaPeckOPC",
    "PeckOPC",
    "LiaoWhitmanOPC",
    "SkemptonOPC",
    "DilatancyCorrection",
    "DilatancyCorrection2",
    "OPCType",
    "create_overburden_pressure_correction",
    "correct_spt_n_value",
]


class SPTDesignMethod(AbstractStrEnum):
    """Enumeration of Standard Penetration Test (SPT) design methods.

    Each member represents a different method for interpreting SPT
    results in geotechnical design calculations.
    """

    MINIMUM = "min"
    """Use the minimum SPT value."""

    AVERAGE = "avg"
    """Use the average SPT value."""

    WEIGHTED = "wgt"
    """Use the weighted SPT value."""


class SPT:
    """SPT Design Calculations.

    Due to uncertainty in field procedure in standard penetration test
    and also to consider all the N-value in the influence zone of a
    foundation, a method was suggested to calculate the design N-value
    which should be used in calculating the allowable bearing capacity
    of shallow foundation rather than using a particular N-value. All
    the N-value from the influence zone is taken under consideration by
    giving the highest weightage to the closest N-value from the base.
    """

    def __init__(
        self,
        corrected_spt_n_values: Sequence[float],
        method: SPTDesignMethod = "wgt",
    ):
        """
        :param corrected_spt_n_values: Corrected SPT N-values within the
                                       foundation influence zone.
        """
        self.corrected_spt_n_values = corrected_spt_n_values
        self.method = method

    @property
    def corrected_spt_n_values(self) -> Sequence[float]:
        """
        Corrected SPT N-values within the foundation influence zone.
        """
        return self._corrected_spt_n_values

    @corrected_spt_n_values.setter
    @validate_params
    def corrected_spt_n_values(
        self,
        corrected_spt_n_values: Annotated[
            Sequence[float],
            MustHaveLengthGreaterThan(1),
            MustHaveValuesBetween(min_value=1.0, max_value=100.0),
        ],
    ) -> None:
        self._corrected_spt_n_values = corrected_spt_n_values

    @property
    def method(self):
        return self._method

    @method.setter
    @validate_params
    def method(self, method: Annotated[str, MustBeMemberOf(SPTDesignMethod)]):
        self._method = method

    @staticmethod
    def _avg_spt_n_design(vals) -> float:
        return mean(vals)

    @staticmethod
    def _min_spt_n_design(vals):
        return min(vals)

    @staticmethod
    def _wgt_spt_n_design(vals):

        total_wgted_spt = 0.0
        total_wgt = 0.0

        for i, corr_spt_n_val in enumerate(vals, start=1):
            wgt = 1 / i**2
            total_wgted_spt += wgt * corr_spt_n_val
            total_wgt += wgt

        return total_wgted_spt / total_wgt

    @round_(ndigits=1)
    def n_design(self):
        r"""Calculates the SPT N-design within the foundation influence
        zone.

        If `method="min"`, it returns the minimum N-value within the
        foundation influence zone as the SPT N-design value. This
        approach was suggested by `Terzaghi & Peck (1948)`.

        if `method="avg"`, it returns the average N-value within the
        foundation influence zone as the SPT N-design value.

        if `method="wgt"`, it returns the weighted average N-value
        within the foundation influence zone as the SPT N-design value.

        $$
        N_{design} = \dfrac{\sum_{i=1}^{n} \frac{N_i}{i^2}}
                      {\sum_{i=1}^{n}\frac{1}{i^2}}
        $$
        """
        if self.method == "min":
            _n_design = self._min_spt_n_design
        elif self.method == "avg":
            _n_design = self._avg_spt_n_design
        else:
            _n_design = self._wgt_spt_n_design

        return _n_design(self.corrected_spt_n_values)


class HammerType(AbstractStrEnum):
    """Enumeration of hammer types used in geotechnical testing.

    Each member represents a different type of hammer used for
    Standard Penetration Test (SPT) or other soil testing methods.
    """

    AUTOMATIC = enum.auto()
    """Automatic hammer."""

    DONUT_1 = enum.auto()
    """Donut-type hammer, variant 1."""

    DONUT_2 = enum.auto()
    """Donut-type hammer, variant 2."""

    SAFETY = enum.auto()
    """Safety hammer."""

    DROP = enum.auto()
    """Drop hammer."""

    PIN = enum.auto()
    """Pin-type hammer."""


class SamplerType(AbstractStrEnum):
    """Enumeration of soil sampler types.

    Each member represents a different type of sampler used in
    Standard Penetration Tests (SPT) or other geotechnical sampling
    methods.
    """

    STANDARD = enum.auto()
    """Standard sampler."""

    NON_STANDARD = enum.auto()
    """Non-standard sampler."""

    LINER_4_DENSE_SAND_AND_CLAY = enum.auto()
    """Liner sampler for dense sand and clay."""

    LINER_4_LOOSE_SAND = enum.auto()
    """Liner sampler for loose sand."""


class EnergyCorrection:
    r"""SPT N-value standardized for field procedures.

    On the basis of field observations, it appears reasonable to
    standardize the field SPT N-value as a function of the input driving
    energy and its dissipation around the sampler around the surrounding
    soil. The variations in testing procedures may be at least partially
    compensated by converting the measured N-value to $N_{60}$
    assuming 60% hammer energy being transferred to the tip of the
    standard split spoon.
    """

    _HAMMER_EFFICIENCY_FACTORS = {
        HammerType.AUTOMATIC: 0.70,
        HammerType.DONUT_1: 0.60,
        HammerType.DONUT_2: 0.50,
        HammerType.SAFETY: 0.55,
        HammerType.DROP: 0.45,
        HammerType.PIN: 0.45,
    }

    _SAMPLER_CORRECTION_FACTORS = {
        SamplerType.STANDARD: 1.00,
        SamplerType.NON_STANDARD: 1.20,
        SamplerType.LINER_4_DENSE_SAND_AND_CLAY: 0.8,
        SamplerType.LINER_4_LOOSE_SAND: 0.9,
    }

    def __init__(
        self,
        recorded_spt_n_value: int,
        *,
        energy_percentage=0.6,
        borehole_diameter=65.0,
        rod_length=3.0,
        hammer_type=HammerType.DONUT_1,
        sampler_type=SamplerType.STANDARD,
    ):
        """
        :param recorded_spt_n_value: Recorded SPT N-value from field.

        :param energy_percentage: Energy percentage reaching the tip of
                                  the sampler.
        :param borehole_diameter: Borehole diameter (mm).
        :param rod_length: Length of SPT rod, defaults to 3.0 (m).
        :param hammer_type: Hammer type.
        :param sampler_type: Sampler type.
        """
        self.recorded_spt_n_value = int(recorded_spt_n_value)
        self.energy_percentage = energy_percentage
        self.borehole_diameter = borehole_diameter
        self.rod_length = rod_length
        self.hammer_type = hammer_type
        self.sampler_type = sampler_type

    @property
    def recorded_spt_n_value(self) -> int:
        """Recorded SPT N-value from field."""
        return self._recorded_spt_value

    @recorded_spt_n_value.setter
    @validate_params
    def recorded_spt_n_value(
        self,
        recorded_spt_n_value: Annotated[int, MustBeBetween(min_value=0, max_value=100)],
    ) -> None:
        self._recorded_spt_value = recorded_spt_n_value

    @property
    def energy_percentage(self) -> float:
        """Energy percentage reaching the tip of the sampler."""
        return self._energy_percentage

    @energy_percentage.setter
    @validate_params
    def energy_percentage(
        self,
        energy_percentage: Annotated[
            float, MustBeBetween(min_value=0.0, max_value=1.0)
        ],
    ) -> None:
        self._energy_percentage = energy_percentage

    @property
    def borehole_diameter(self) -> float:
        """Borehole diameter (mm)."""
        return self._borehole_diameter

    @borehole_diameter.setter
    @validate_params
    def borehole_diameter(
        self,
        borehole_diameter: Annotated[
            float, MustBeBetween(min_value=65.0, max_value=200.0)
        ],
    ) -> None:
        self._borehole_diameter = borehole_diameter

    @property
    def rod_length(self) -> float:
        """Length of SPT rod."""
        return self._rod_length

    @rod_length.setter
    @validate_params
    def rod_length(self, rod_length: Annotated[float, MustBePositive()]):
        self._rod_length = rod_length

    @property
    def hammer_type(self) -> HammerType:
        return self._hammer_type

    @hammer_type.setter
    @validate_params
    def hammer_type(
        self,
        hammer_type: Annotated[HammerType, MustBeMemberOf(HammerType)],
    ):
        self._hammer_type = hammer_type

    @property
    def sampler_type(self) -> SamplerType:
        return self._sampler_type

    @sampler_type.setter
    @validate_params
    def sampler_type(
        self, sampler_type: Annotated[SamplerType, MustBeMemberOf(SamplerType)]
    ):
        self._sampler_type = sampler_type

    @property
    def hammer_efficiency(self) -> float:
        """Hammer efficiency correction factor."""
        return self._HAMMER_EFFICIENCY_FACTORS[self.hammer_type]

    @property
    def borehole_diameter_correction(self) -> float:
        """Borehole diameter correction factor."""
        if 65 <= self.borehole_diameter <= 115:
            corr = 1.00
        elif 115 < self.borehole_diameter <= 150:
            corr = 1.05
        else:
            corr = 1.15
        return corr

    @property
    def sampler_correction(self) -> float:
        """Sampler correction factor."""
        return self._SAMPLER_CORRECTION_FACTORS[self.sampler_type]

    @property
    def rod_length_correction(self) -> float:
        """Rod length correction factor."""
        if 3.0 <= self.rod_length <= 4.0:
            corr = 0.75
        elif 4.0 < self.rod_length <= 6.0:
            corr = 0.85
        elif 6.0 < self.rod_length <= 10.0:
            corr = 0.95
        else:
            corr = 1.00
        return corr

    def correction(self) -> float:
        r"""Energy correction factor.

        $$
        N_{ENERGY} = \dfrac{E_H \cdot C_B \cdot C_S
                     \cdot C_R \cdot N}{ENERGY}
        $$

        `ENERGY`: 0.6, 0.55, etc
        """
        numerator = (
            self.hammer_efficiency
            * self.borehole_diameter_correction
            * self.sampler_correction
            * self.rod_length_correction
        )
        return numerator / self.energy_percentage

    @round_(ndigits=1)
    def standardized_spt_n_value(self) -> float:
        """Standardized SPT N-value."""
        return self.correction() * self.recorded_spt_n_value


class OPC(ABC):
    """Base class for Overburden Pressure Correction (OPC)."""

    def __init__(self, std_spt_n_value: float, eop: float):
        """
        :param std_spt_n_value: SPT N-value standardized for field
                                procedures.
        :param eop: Effective overburden pressure ($kPa$).
        """
        self.std_spt_n_value = std_spt_n_value
        self.eop = eop

    @property
    def eop(self) -> float:
        """Effective overburden pressure ($kPa$)."""
        return self._eop

    @eop.setter
    @validate_params
    def eop(self, eop: Annotated[float, MustBeNonNegative()]):
        """Effective overburden pressure ($kPa$)."""
        self._eop = eop

    @property
    def std_spt_n_value(self) -> float:
        """SPT N-value standardized for field procedures."""
        return self._std_spt_n_value

    @std_spt_n_value.setter
    @validate_params
    def std_spt_n_value(
        self,
        std_spt_n_value: Annotated[
            float, MustBeBetween(min_value=0.0, max_value=100.0)
        ],
    ):
        self._std_spt_n_value = std_spt_n_value

    @round_(ndigits=1)
    def corrected_spt_n_value(self) -> float:
        r"""Corrected SPT N-value."""
        # Corrected SPT should not be more
        # than 2 times the Standardized SPT
        correction = min(self.correction(), 2.0)
        return correction * self.std_spt_n_value

    @abstractmethod
    def correction(self) -> float:
        raise NotImplementedError


class GibbsHoltzOPC(OPC):
    """Overburden Pressure Correction according to
    `Gibbs & Holtz (1957)`.
    """

    @property
    def eop(self) -> float:
        """Effective overburden pressure ($kPa$)."""
        return self._eop

    @eop.setter
    @validate_params
    def eop(
        self,
        eop: Annotated[float, MustBeBetween(min_value=0.0, max_value=280.0)],
    ):
        self._eop = eop

    def correction(self) -> float:
        r"""SPT Correction."""
        corr = 350.0 / (self.eop + 70.0)
        return corr / 2.0 if corr > 2.0 else corr


class BazaraaPeckOPC(OPC):
    """Overburden Pressure Correction according to `Bazaraa (1967)`, and
    also by `Peck and Bazaraa (1969)`.
    """

    #: Maximum effective overburden pressure (:math:`kPa`).
    STD_PRESSURE: Final = 71.8

    def correction(self) -> float:
        r"""SPT Correction."""
        if isclose(self.eop, self.STD_PRESSURE, rel_tol=0.01):
            corr = 1.0
        elif self.eop < self.STD_PRESSURE:
            corr = 4.0 / (1.0 + 0.0418 * self.eop)
        else:
            corr = 4.0 / (3.25 + 0.0104 * self.eop)
        return corr


class PeckOPC(OPC):
    """Overburden Pressure Correction according to
    `Peck et al. (1974)`.
    """

    @property
    def eop(self) -> float:
        """Effective overburden pressure ($kPa$)."""
        return self._eop

    @eop.setter
    @validate_params
    def eop(self, eop: Annotated[float, MustBeGreaterThanOrEqual(24.0)]):
        self._eop = eop

    def correction(self) -> float:
        r"""SPT Correction."""
        return 0.77 * log10(2000.0 / self.eop)


class LiaoWhitmanOPC(OPC):
    """Overburden Pressure Correction according to
    `Liao & Whitman (1986)`.
    """

    def correction(self) -> float:
        r"""SPT Correction."""
        return sqrt(100.0 / self.eop)


class SkemptonOPC(OPC):
    """Overburden Pressure Correction according to `Skempton (1986)`."""

    def correction(self) -> float:
        r"""SPT Correction."""
        return 2.0 / (1.0 + 0.01044 * self.eop)


class DilatancyCorrection:
    """Dilatancy SPT Correction according to `Terzaghi & Peck (1948)`.

    For coarse sand, this correction is not required. In applying this
    correction, overburden pressure correction is applied first and then
    dilatancy correction is applied.
    """

    def __init__(self, corr_spt_n_value: float):
        """
        :param corr_spt_n_value: SPT N-value standardized for field
                                 procedures and/or corrected for
                                 overburden pressure.
        """
        self.corr_spt_n_value = corr_spt_n_value

    @property
    def corr_spt_n_value(self) -> float:
        """SPT N-value standardized for field procedures and/or corrected
        for overburden pressure.
        """
        return self._corr_spt_n_value

    @corr_spt_n_value.setter
    @validate_params
    def corr_spt_n_value(
        self,
        corr_spt_n_value: Annotated[
            float, MustBeBetween(min_value=0.0, max_value=100.0)
        ],
    ):
        self._corr_spt_n_value = corr_spt_n_value

    @round_(ndigits=1)
    def corrected_spt_n_value(self) -> float:
        r"""Corrected SPT N-value for presence of water."""
        if self.corr_spt_n_value <= 15.0:
            return self.corr_spt_n_value
        return 15.0 + 0.5 * (self.corr_spt_n_value - 15.0)


class DilatancyCorrection2(DilatancyCorrection):
    """Correction factor for groundwater level."""

    def __init__(self, corr_spt_n_value: float, foundation_size: Foundation):
        super().__init__(corr_spt_n_value)
        self.foundation_size = foundation_size

    def correction(self) -> float:
        water_lvl = self.foundation_size.ground_water_level
        f_width = self.foundation_size.width
        f_depth = self.foundation_size.depth
        correction = 1.0

        if not isinf(water_lvl):
            if (d1 := water_lvl - f_depth) <= f_width or water_lvl < f_width:
                correction = 0.5 + (d1 / (2 * (f_depth + f_width)))

        return correction

    def corrected_spt_n_value(self) -> float:
        """Corrected SPT N-value for presence of water."""
        correction = min(self.correction(), 2.0)
        return correction * self.corr_spt_n_value


class OPCType(AbstractStrEnum):
    """Enumeration of overburden pressure correction (OPC) methods.

    Each member represents a method used to correct SPT results
    for the effects of overburden pressure in geotechnical design.
    """

    GIBBS = enum.auto()
    """Gibbs method for overburden pressure correction."""

    BAZARAA = enum.auto()
    """Bazaraa method for overburden pressure correction."""

    PECK = enum.auto()
    """Peck method for overburden pressure correction."""

    LIAO = enum.auto()
    """Liao method for overburden pressure correction."""

    SKEMPTON = enum.auto()
    """Skempton method for overburden pressure correction."""


class DilatancyCorrType(AbstractStrEnum):
    NON_WATER_AWARE = enum.auto()
    WATER_AWARE = enum.auto()


_opc_methods: Final = {
    OPCType.GIBBS: GibbsHoltzOPC,
    OPCType.BAZARAA: BazaraaPeckOPC,
    OPCType.PECK: PeckOPC,
    OPCType.LIAO: LiaoWhitmanOPC,
    OPCType.SKEMPTON: SkemptonOPC,
}


@validate_params
def correct_spt_n_value(
    recorded_spt_n_value: int,
    *,
    eop,
    energy_percentage=0.6,
    borehole_diameter=65.0,
    rod_length=3.0,
    hammer_type=HammerType.DONUT_1,
    sampler_type=SamplerType.STANDARD,
    opc_method: OPCType = "gibbs",
    dilatancy_corr_method: Annotated[
        Optional[DilatancyCorrType], MustBeMemberOf(DilatancyCorrType)
    ] = None,
    foundation_size: Annotated[
        Foundation,
        DependsOn(dilatancy_corr_method=DilatancyCorrType.WATER_AWARE),
    ] = None,
) -> float:
    """SPT N-value correction for overburden pressure and groundwater
    level.

    :param recorded_spt_n_value: Recorded SPT N-value from field.
    :param eop: Effective overburden pressure ($kPa$).
    :param energy_percentage: Energy percentage reaching the tip of
                              the sampler.
    :param borehole_diameter: Borehole diameter (mm).
    :param rod_length: Length of SPT rod, defaults to 3.0 (m).
    :param hammer_type: Hammer type.
    :param sampler_type: Sampler type.
    :param opc_method: Overburden pressure correction method.
    :param dilatancy_corr_method: Dilatancy correction method.
    :param foundation_size: Foundation size.
    """
    energy_correction = EnergyCorrection(
        recorded_spt_n_value,
        energy_percentage=energy_percentage,
        borehole_diameter=borehole_diameter,
        rod_length=rod_length,
        hammer_type=hammer_type,
        sampler_type=sampler_type,
    )
    std_spt_n_value = energy_correction.standardized_spt_n_value()
    opc_corr = create_overburden_pressure_correction(
        std_spt_n_value=std_spt_n_value, eop=eop, opc_method=opc_method
    )
    corr_spt_n_value = opc_corr.corrected_spt_n_value()

    if dilatancy_corr_method is not None:
        if dilatancy_corr_method == DilatancyCorrType.NON_WATER_AWARE:
            dil_corr = DilatancyCorrection(corr_spt_n_value=corr_spt_n_value)
            corr_spt_n_value = dil_corr.corrected_spt_n_value()
        else:
            dil_corr = DilatancyCorrection2(
                corr_spt_n_value=corr_spt_n_value,
                foundation_size=foundation_size,
            )
            corr_spt_n_value = dil_corr.corrected_spt_n_value()

    return corr_spt_n_value


@validate_params
def create_overburden_pressure_correction(
    std_spt_n_value: float,
    eop: float,
    opc_method: Annotated[OPCType | str, MustBeMemberOf(OPCType)] = "gibbs",
):
    """A factory function that encapsulates the creation of overburden
    pressure correction.

    :param std_spt_n_value: SPT N-value standardized for field
                            procedures.
    :param eop: Effective overburden pressure ($kPa$).
    :param opc_method: Overburden Pressure Correction type to apply.
    """
    opc_method = OPCType(opc_method)
    opc_class = _opc_methods[opc_method]
    opc_corr = opc_class(std_spt_n_value=std_spt_n_value, eop=eop)
    return opc_corr



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Annotated, Optional

from func_validator import MustBeLessThanOrEqual, MustBeNonNegative, validate_params

from geolysis.foundation import Foundation
from geolysis.utils import round_


@dataclass
class AllowableBearingCapacityResult:
    allowable_bearing_capacity: float
    depth_factor: float
    water_correction_factor: Optional[float] = None


class AllowableBearingCapacity(ABC):
    #: Maximum tolerable foundation settlement (mm).
    MAX_TOL_SETTLEMENT = 25.4

    def __init__(
        self,
        corrected_spt_n_value: float,
        tol_settlement: float,
        foundation_size: Foundation,
    ) -> None:
        self.corrected_spt_n_value = corrected_spt_n_value
        self.tol_settlement = tol_settlement
        self.foundation_size = foundation_size

    @property
    def corrected_spt_n_value(self) -> float:
        """Statistical average of corrected SPT N-value."""
        return self._corrected_spt_n_value

    @corrected_spt_n_value.setter
    @validate_params
    def corrected_spt_n_value(
        self,
        corrected_spt_n_value: Annotated[float, MustBeNonNegative()],
    ):
        self._corrected_spt_n_value = corrected_spt_n_value

    @property
    def tol_settlement(self) -> float:
        """Tolerable settlement of foundation (mm)."""
        return self._tol_settlement

    @tol_settlement.setter
    @validate_params
    def tol_settlement(
        self,
        tol_settlement: Annotated[
            float, MustBeNonNegative(), MustBeLessThanOrEqual(25.4)
        ],
    ):
        self._tol_settlement = tol_settlement

    def _sr(self) -> float:
        """Calculate the settlement ratio."""
        return self.tol_settlement / self.MAX_TOL_SETTLEMENT

    def _fd(self) -> float:
        """Calculate the depth factor."""
        depth = self.foundation_size.depth
        width = self.foundation_size.width
        return min(1.0 + 0.33 * depth / width, 1.33)

    def bearing_capacity_results(self) -> AllowableBearingCapacityResult:
        """Return bearing capacity results with intermediate calculations.

        !!! info "Added in v0.11.0"
        """
        return AllowableBearingCapacityResult(
            allowable_bearing_capacity=self.allowable_bearing_capacity(),
            depth_factor=self._fd(),
        )

    @round_(ndigits=1)
    def allowable_bearing_capacity(self) -> float:
        """Calculates the allowable bearing capacity.

        !!! info "Added in v0.12.0"
        """
        return self._bearing_capacity()

    @round_(ndigits=1)
    def allowable_applied_load(self) -> float:
        """Calculate the allowable applied load on the foundation."""
        return self._bearing_capacity() * self.foundation_size.foundation_area()

    @abstractmethod
    def _bearing_capacity(self) -> float: ...



import enum
from typing import Annotated, Optional

from func_validator import MustBeMemberOf, validate_params

from geolysis.foundation import FoundationType, Shape, create_foundation
from geolysis.utils import AbstractStrEnum, inf

from ._core import AllowableBearingCapacity
from .bowles_abc import BowlesABC4MatFoundation, BowlesABC4PadFoundation
from .meyerhof_abc import MeyerhofABC4MatFoundation, MeyerhofABC4PadFoundation
from .terzaghi_abc import TerzaghiABC4MatFoundation, TerzaghiABC4PadFoundation


class ABCType(AbstractStrEnum):
    """Enumeration of allowable bearing capacity calculation methods.

    Each member represents a different method for determining
    the allowable bearing capacity of soil.
    """

    BOWLES = enum.auto()
    """Bowles's method for calculating allowable bearing capacity"""

    MEYERHOF = enum.auto()
    """Meyerhof's method for calculating allowable bearing capacity"""

    TERZAGHI = enum.auto()
    """Terzaghi's method for calculating allowable bearing capacity"""


abc_classes = {
    ABCType.BOWLES: {
        FoundationType.PAD: BowlesABC4PadFoundation,
        FoundationType.MAT: BowlesABC4MatFoundation,
    },
    ABCType.MEYERHOF: {
        FoundationType.PAD: MeyerhofABC4PadFoundation,
        FoundationType.MAT: MeyerhofABC4MatFoundation,
    },
    ABCType.TERZAGHI: {
        FoundationType.PAD: TerzaghiABC4PadFoundation,
        FoundationType.MAT: TerzaghiABC4MatFoundation,
    },
}


@validate_params
def create_abc_4_cohesionless_soils(
    corrected_spt_n_value: float,
    tol_settlement: float,
    depth: float,
    width: float,
    length: Optional[float] = None,
    eccentricity: float = 0.0,
    ground_water_level: float = inf,
    shape: Shape | str = "square",
    foundation_type: FoundationType | str = "pad",
    abc_type: Annotated[ABCType | str, MustBeMemberOf(ABCType)] = "bowles",
) -> AllowableBearingCapacity:
    r"""A factory function that encapsulate the creation of  allowable
     bearing capacities.

    :param corrected_spt_n_value: The corrected SPT N-value.
    :param tol_settlement: Tolerable settlement of foundation (mm).
    :param depth: Depth of foundation (m).
    :param width: Width of foundation footing (m).
    :param length: Length of foundation footing (m).
    :param eccentricity: The deviation of the foundation load from the
                         center of gravity of the foundation footing (m).
    :param ground_water_level: Depth of water below ground level (m).
    :param shape: Shape of foundation footing
    :param foundation_type: Type of foundation.
    :param abc_type: Type of allowable bearing capacity calculation to
                     apply.
    """
    abc_type = ABCType(abc_type)
    foundation_type = FoundationType(foundation_type)

    # exception from create_foundation will automaatically propagate
    # no need to catch and handle it.
    fnd_size = create_foundation(
        depth=depth,
        width=width,
        length=length,
        eccentricity=eccentricity,
        ground_water_level=ground_water_level,
        foundation_type=foundation_type,
        shape=shape,
    )
    abc_class = abc_classes[abc_type][foundation_type]

    return abc_class(
        corrected_spt_n_value=corrected_spt_n_value,
        tol_settlement=tol_settlement,
        foundation_size=fnd_size,
    )



import enum
import functools
from typing import Callable

from . import math as m
from .math import *

__all__ = ["AbstractStrEnum", "round_"] + m.__all__


class StrEnumMeta(enum.EnumMeta):
    def __contains__(cls, item) -> bool:
        if isinstance(item, (str, cls)):
            return str(item) in (member.value for member in cls)
        return NotImplemented

    def __repr__(cls) -> str:
        return str([member.value for member in cls])


class AbstractStrEnum(enum.StrEnum, metaclass=StrEnumMeta):
    """An abstract string enumeration class that inherits from StrEnum.

    This class can be used as a base class for creating string
    enumerations.
    """


def round_(ndigits: int) -> Callable:
    """A decorator that rounds the result of a callable to a specified
    number of decimal places.

    The returned value of the callable should support the `__round__`
    dunder method and should be a numeric value.

    TypeError is raised when `ndigits` is not an int.
    """

    if not isinstance(ndigits, int):
        raise TypeError("ndigits must be an int")

    def dec(fn) -> Callable[..., float]:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs) -> float:
            res = fn(*args, **kwargs)
            return round(res, ndigits=ndigits)

        return wrapper

    return dec



from geolysis.foundation import Foundation
from geolysis.utils import isinf, round_

from ._core import AllowableBearingCapacity, AllowableBearingCapacityResult


class TerzaghiABC4PadFoundation(AllowableBearingCapacity):
    """Allowable bearing capacity for pad foundation on cohesionless
    soils according to `Terzaghi & Peck (1948)`.

    See [implementation](../formulas/allowable-bearing-capacity.md/#terzaghi-bearing-capacity-for-pad-foundation)
    for more details on bearing capacity equation used.

    """

    def __init__(
        self,
        corrected_spt_n_value: float,
        tol_settlement: float,
        foundation_size: Foundation,
    ) -> None:
        """
        :param corrected_spt_n_value: Lowest (or average) uncorrected
                                      SPT N-value (60% energy) within
                                      the foundation influence zone i.e.
                                      $D_f$ to $D_f + 2B$
        :param tol_settlement: Tolerable settlement of foundation (mm).
        :param foundation_size: Size of the foundation.
        """
        super().__init__(
            corrected_spt_n_value=corrected_spt_n_value,
            tol_settlement=tol_settlement,
            foundation_size=foundation_size,
        )

    def _fd(self) -> float:
        """Calculate the depth factor."""
        depth = self.foundation_size.depth
        width = self.foundation_size.width

        return min(1.0 + 0.25 * depth / width, 1.25)

    def _cw(self):
        """Calculate the water correction factor."""
        depth = self.foundation_size.depth
        width = self.foundation_size.width
        water_level = self.foundation_size.ground_water_level

        if isinf(water_level):
            return 2.0

        if water_level <= depth:
            cw = 2.0 - depth / (2.0 * width)
        else:
            cw = 2.0 - water_level / (2.0 * width)

        return min(cw, 2.0)

    @round_(ndigits=2)
    def _bearing_capacity(self):
        """
        Calculates the allowable bearing capacity of the pad foundation.
        """
        n_corr = self.corrected_spt_n_value
        width = self.foundation_size.width

        if width <= 1.2:
            return 12 * n_corr * (1 / (self._cw() * self._fd())) * self._sr()

        return (
            8
            * n_corr
            * ((3.28 * width + 1) / (3.28 * width)) ** 2
            * (1 / (self._cw() * self._fd()))
            * self._sr()
        )

    def bearing_capacity_results(self) -> AllowableBearingCapacityResult:
        res = super().bearing_capacity_results()
        res.water_correction_factor = self._cw()
        return res


class TerzaghiABC4MatFoundation(TerzaghiABC4PadFoundation):
    """Allowable bearing capacity for mat foundation on cohesionless
    soils according to `Terzaghi & Peck (1948)`.

    See [implementation](../formulas/allowable-bearing-capacity.md/#terzaghi-bearing-capacity-for-mat-foundation)
    for more details on bearing capacity equation used.

    """

    @round_(ndigits=2)
    def _bearing_capacity(self):
        """
        Calculates the allowable bearing capacity of the mat foundation.
        """
        n_corr = self.corrected_spt_n_value
        return 8 * n_corr * (1 / (self._cw() * self._fd())) * self._sr()



import enum
from abc import ABC, abstractmethod
from typing import Annotated, Optional, TypeVar

from func_validator import (
    DependsOn,
    MustBeBetween,
    MustBeMemberOf,
    MustBeNonNegative,
    MustBePositive,
    validate_params,
)

from .utils import AbstractStrEnum, inf, isclose, isinf, pi, round_

__all__ = [
    "create_foundation",
    "Foundation",
    "FoundationType",
    "Shape",
    "StripFooting",
    "CircularFooting",
    "SquareFooting",
    "RectangularFooting",
]

T = TypeVar("T")


class Shape(AbstractStrEnum):
    """Enumeration of foundation shapes.

    Each member represents a standard geometric shape commonly used
    in foundation design, which can affect bearing capacity and
    settlement calculations.
    """

    STRIP = enum.auto()
    """Strip (or continuous) foundation, typically long and narrow."""

    CIRCLE = enum.auto()
    """Circular foundation, often used for columns or piers."""

    SQUARE = enum.auto()
    """Square foundation, commonly used for isolated footings."""

    RECTANGLE = enum.auto()
    """Rectangular foundation, used when length and width differ significantly."""


class FoundationType(AbstractStrEnum):
    """
    Enumeration of foundation types.

    Each member represents a common type of foundation used in
    geotechnical engineering. Some members have aliases for
    alternative naming conventions.
    """

    PAD = enum.auto()
    """Pad foundation, a small, isolated footing supporting a single column."""

    ISOLATED = PAD
    """Alias for PAD foundation (isolated footing)."""

    MAT = enum.auto()
    """Mat foundation, a large, continuous slab supporting multiple columns or 
    walls.
    """

    RAFT = MAT
    """Alias for MAT foundation (raft foundation)."""


class FootingSize(ABC):
    _SHAPE: Shape

    @property
    @abstractmethod
    def width(self) -> float:
        raise NotImplementedError

    @width.setter
    def width(self, width: float):
        raise NotImplementedError

    @property
    @abstractmethod
    def length(self) -> float:
        raise NotImplementedError

    @length.setter
    def length(self, length: float):
        raise NotImplementedError

    def area(self) -> float:
        raise NotImplementedError

    @property
    def shape(self) -> Shape:
        """Return the shape of the foundation footing."""
        return self._SHAPE


class StripFooting(FootingSize):
    """A class representation of strip footing."""

    _SHAPE = Shape.STRIP

    def __init__(self, width: float, length: float = inf):
        """
        :param width: Width of foundation footing (m).
        :param length: Length of foundation footing (m).
        """
        self.width = width
        self.length = length

    @property
    def width(self) -> float:
        """Width of foundation footing (m)."""
        return self._width

    @width.setter
    def width(self, width: float) -> None:
        self._width = width

    @property
    def length(self) -> float:
        """Length of foundation footing (m)."""
        return self._length

    @length.setter
    def length(self, length: float) -> None:
        self._length = length

    def area(self) -> float:
        """Area of strip footing ($m \text{or} m^2$)."""
        if isinf(self.length):
            return self.width
        return self.width * self.length


class CircularFooting(FootingSize):
    """A class representation of circular footing.

    !!! note

        The `width` and `length` properties refer to the diameter of the
        circular footing. This is to make it compatible with the protocol
        square and rectangular footing follow.
    """

    _SHAPE = Shape.CIRCLE

    def __init__(self, diameter: float):
        """
        :param diameter: Diameter of foundation footing (m).
        """
        self.diameter = diameter

    @property
    def diameter(self) -> float:
        """Diameter of foundation footing (m)."""
        return self._diameter

    @diameter.setter
    def diameter(self, diameter: float) -> None:
        self._diameter = diameter

    @property
    def width(self):
        """Diameter of foundation footing (m)."""
        return self.diameter

    # Not checking for positive as diameter setter already does that
    @width.setter
    def width(self, width: float):
        self.diameter = width

    # Not checking for positive as diameter setter already does that
    @property
    def length(self):
        """Diameter of foundation footing (m)."""
        return self.diameter

    @length.setter
    def length(self, length: float):
        self.diameter = length

    def area(self) -> float:
        """Area of circular footing ($m^2$)."""
        return pi * self.diameter**2 / 4


class SquareFooting(FootingSize):
    """A class representation of square footing."""

    _SHAPE = Shape.SQUARE

    def __init__(self, width: float):
        """
        :param float width: Width of foundation footing (m).
        """
        self.width = width

    @property
    def width(self):
        """Width of foundation footing (m)."""
        return self._width

    @width.setter
    def width(self, width: float) -> None:
        self._width = width

    @property
    def length(self):
        """Width of foundation footing (m)."""
        return self.width

    # Not checking for positive as width setter already does that
    @length.setter
    def length(self, length: float):
        self.width = length

    def area(self) -> float:
        """Area of square footing ($m^2$)."""
        return self.width**2


class RectangularFooting(FootingSize):
    """A class representation of rectangular footing."""

    _SHAPE = Shape.RECTANGLE

    def __init__(self, width: float, length: float):
        """
        :param width: Width of foundation footing (m).
        :param length: Length of foundation footing (m).
        """
        self.width = width
        self.length = length

    @property
    def width(self) -> float:
        """Width of foundation footing (m)."""
        return self._width

    @width.setter
    def width(self, width: float) -> None:
        self._width = width

    @property
    def length(self) -> float:
        """Length of foundation footing (m)."""
        return self._length

    @length.setter
    def length(self, length: float) -> None:
        self._length = length

    def area(self) -> float:
        """Area of rectangular footing ($m^2$)."""
        return self.width * self.length


class Foundation:
    """A simple class representing a foundation structure."""

    def __init__(
        self,
        depth: float,
        footing_size: FootingSize,
        eccentricity: float = 0.0,
        load_angle: float = 0.0,
        ground_water_level: Optional[float] = inf,
        foundation_type: FoundationType = FoundationType.PAD,
    ) -> None:
        r"""
        :param depth: Depth of foundation (m).
        :param footing_size: Represents the size of the foundation footing.
        :param eccentricity: The deviation of the foundation load from
                             the center of gravity of the foundation
                             footing (m).
        :param load_angle: Inclination of the applied load with the
                           vertical ($\alpha^{\circ}$)
        :param ground_water_level: Depth of the water below ground level
                                   (m).
        :param foundation_type: Type of foundation.
        """
        self.depth = depth
        self.footing_size = footing_size
        self.eccentricity = eccentricity
        self.load_angle = load_angle
        self.ground_water_level = ground_water_level
        self.foundation_type = foundation_type

    @property
    def depth(self) -> float:
        """Depth of foundation (m)."""
        return self._depth

    @depth.setter
    @validate_params
    def depth(self, depth: Annotated[float, MustBePositive()]) -> None:
        self._depth = depth

    @property
    def width(self) -> float:
        """Width of foundation footing (m)."""
        return self.footing_size.width

    @width.setter
    def width(self, width: Annotated[float, MustBePositive()]):
        self.footing_size.width = width

    @property
    def length(self) -> float:
        """Length of foundation footing (m)."""
        return self.footing_size.length

    @length.setter
    def length(self, length: Annotated[float, MustBePositive()]):
        self.footing_size.length = length

    @property
    def footing_shape(self) -> Shape:
        """Shape of foundation footing."""
        return self.footing_size.shape

    @property
    def eccentricity(self) -> float:
        """The deviation of the foundation load from the center of
        gravity of the foundation footing (m).
        """
        return self._eccentricity

    @eccentricity.setter
    @validate_params
    def eccentricity(self, eccentricity: Annotated[float, MustBeNonNegative()]) -> None:
        self._eccentricity = eccentricity

    @property
    def load_angle(self) -> float:
        """Inclination of the applied load with the  vertical."""
        return self._load_angle

    @load_angle.setter
    @validate_params
    def load_angle(
        self,
        load_angle: Annotated[float, MustBeBetween(min_value=0.0, max_value=90.0)],
    ) -> None:
        self._load_angle = load_angle

    @property
    def ground_water_level(self) -> Optional[float]:
        """Depth of the water below ground level (m)."""
        return self._ground_water_level

    @ground_water_level.setter
    @validate_params
    def ground_water_level(
        self,
        ground_water_level: Annotated[float, MustBePositive()],
    ):
        self._ground_water_level = ground_water_level

    @property
    def foundation_type(self) -> FoundationType:
        """Type of foundation."""
        return self._foundation_type

    @foundation_type.setter
    @validate_params
    def foundation_type(
        self,
        foundation_type: Annotated[FoundationType, MustBeMemberOf(FoundationType)],
    ):
        self._foundation_type = foundation_type

    @round_(2)
    def foundation_area(self) -> float:
        """Returns the area of the foundation footing ($m^2$)."""
        return self.footing_size.area()

    @property
    def effective_width(self) -> float:
        """Returns the effective width of the foundation footing (m)."""
        return self.width - 2.0 * self.eccentricity

    def footing_params(self) -> tuple[float, float, Shape]:
        """Returns the `effective_width`, `length`, and `footing_shape`
        of the foundation footing.
        """
        width, length, shape = (
            self.effective_width,
            self.length,
            self.footing_shape,
        )

        if not isclose(width, length) and shape != Shape.STRIP:
            shape = Shape.RECTANGLE

        return width, length, shape


@validate_params
def create_foundation(
    depth: float,
    width: float,
    length: Annotated[float, DependsOn(shape=Shape.RECTANGLE)] = None,
    eccentricity: float = 0.0,
    load_angle: float = 0.0,
    ground_water_level: Optional[float] = inf,
    foundation_type: FoundationType = "pad",
    shape: Annotated[Shape | str, MustBeMemberOf(Shape)] = "square",
) -> Foundation:
    r"""A factory function that encapsulate the creation of a foundation.

    :param depth: Depth of foundation (m).
    :param width: Width of foundation footing. In the case of a circular
                  footing, it refers to the footing diameter (m).
    :param length: Length of foundation footing (m), defaults to None.
    :param eccentricity: The deviation of the foundation load from the
                         center of gravity of the foundation footing (m),
                         defaults to 0.0. This means that the foundation
                         load aligns with the center of gravity of the
                         foundation footing .
    :param load_angle: Inclination of the applied load with the  vertical
                       (:math:`\alpha^{\circ}`).
    :param ground_water_level: Depth of the water below ground level (m)
    :param foundation_type: Type of foundation footing.
    :param shape: Shape of foundation footing

    :raises ValidationError: Raised when length is not provided for a
                             rectangular footing or an invalid shape is
                             provided.
    """
    footing_size: FootingSize
    shape = Shape(str(shape).casefold())

    if shape == Shape.STRIP:
        footing_size = StripFooting(width=width)
    elif shape == Shape.SQUARE:
        footing_size = SquareFooting(width=width)
    elif shape == Shape.CIRCLE:
        footing_size = CircularFooting(diameter=width)
    else:  # RECTANGLE
        footing_size = RectangularFooting(width=width, length=length)

    return Foundation(
        depth=depth,
        eccentricity=eccentricity,
        load_angle=load_angle,
        ground_water_level=ground_water_level,
        foundation_type=foundation_type,
        footing_size=footing_size,
    )



import math
from math import atan, exp, inf, isclose, isinf, isnan, log10, nan, pi, sqrt
from statistics import fmean as mean

__all__ = [
    "atan",
    "inf",
    "isinf",
    "pi",
    "deg2rad",
    "rad2deg",
    "tandeg",
    "cotdeg",
    "sindeg",
    "cosdeg",
    "arctandeg",
    "mean",
    "exp",
    "isclose",
    "log10",
    "sqrt",
    "nan",
    "isnan",
]


def deg2rad(x: float, /) -> float:
    """Convert angle x from degrees to radians."""
    return math.radians(x)


def rad2deg(x: float, /) -> float:
    """Convert angle x from radians to degrees."""
    return math.degrees(x)


def tandeg(x: float, /) -> float:
    """Return the tangent of x (measured in degrees)."""
    return math.tan(deg2rad(x))


def cotdeg(x: float, /) -> float:
    """Return the cotangent of x (measured in degrees)."""
    return 1 / tandeg(x)


def sindeg(x: float, /) -> float:
    """Return the sine of x (measured in degrees)."""
    return math.sin(deg2rad(x))


def cosdeg(x: float, /) -> float:
    """Return the cosine of x (measured in degrees)."""
    return math.cos(deg2rad(x))


def arctandeg(x: float, /) -> float:
    """Return the arc tangent (measured in degrees) of x."""
    return rad2deg(math.atan(x))



from ._cohl import (
    ABCType,
    BowlesABC4MatFoundation,
    BowlesABC4PadFoundation,
    MeyerhofABC4MatFoundation,
    MeyerhofABC4PadFoundation,
    TerzaghiABC4MatFoundation,
    TerzaghiABC4PadFoundation,
    create_abc_4_cohesionless_soils,
)

__all__ = [
    "BowlesABC4PadFoundation",
    "BowlesABC4MatFoundation",
    "MeyerhofABC4PadFoundation",
    "MeyerhofABC4MatFoundation",
    "TerzaghiABC4PadFoundation",
    "TerzaghiABC4MatFoundation",
    "ABCType",
    "create_abc_4_cohesionless_soils",
]



from geolysis.utils import cosdeg, cotdeg, deg2rad, exp, isclose, pi, round_, tandeg

from ._core import UltimateBearingCapacity

__all__ = [
    "TerzaghiUBC4StripFooting",
    "TerzaghiUBC4CircularFooting",
    "TerzaghiUBC4SquareFooting",
    "TerzaghiUBC4RectangularFooting",
]


class TerzaghiBearingCapacityFactors:

    @staticmethod
    @round_(ndigits=2)
    def n_c(friction_angle: float) -> float:
        if isclose(friction_angle, 0.0):
            return 5.7
        return cotdeg(friction_angle) * (
            TerzaghiBearingCapacityFactors.n_q(friction_angle) - 1.0
        )

    @staticmethod
    @round_(ndigits=2)
    def n_q(friction_angle: float) -> float:
        return exp(
            (3.0 * pi / 2.0 - deg2rad(friction_angle)) * tandeg(friction_angle)
        ) / (2.0 * (cosdeg(45.0 + friction_angle / 2.0)) ** 2.0)

    @staticmethod
    @round_(ndigits=2)
    def n_gamma(friction_angle: float) -> float:
        return (TerzaghiBearingCapacityFactors.n_q(friction_angle) - 1.0) * tandeg(
            1.4 * friction_angle
        )


class TerzaghiUltimateBearingCapacity(UltimateBearingCapacity):

    @property
    def n_c(self) -> float:
        r"""Bearing capacity factor $N_c$."""
        return TerzaghiBearingCapacityFactors.n_c(self.friction_angle)

    @property
    def n_q(self) -> float:
        r"""Bearing capacity factor $N_q$."""
        return TerzaghiBearingCapacityFactors.n_q(self.friction_angle)

    @property
    def n_gamma(self) -> float:
        r"""Bearing capacity factor $N_{\gamma}$."""
        return TerzaghiBearingCapacityFactors.n_gamma(self.friction_angle)


class TerzaghiUBC4StripFooting(TerzaghiUltimateBearingCapacity):
    """Ultimate bearing capacity for strip footing according to
    `Terzaghi 1943`.

    See [implementation](../formulas/ultimate-bearing-capacity.md/#terzaghi-bearing-capacity-for-strip-footing)
    for more details on bearing capacity equation used.
    """

    @round_(ndigits=2)
    def _bearing_capacity(self) -> float:
        """Calculates ultimate bearing capacity for strip footing."""
        return (
            self._cohesion_term(1.0)
            + self._surcharge_term()
            + self._embedment_term(0.5)
        )


class TerzaghiUBC4CircularFooting(TerzaghiUltimateBearingCapacity):
    """Ultimate bearing capacity for circular footing according to
    `Terzaghi 1943`.

    See [implementation](../formulas/ultimate-bearing-capacity.md/#terzaghi-bearing-capacity-for-circular-footing)
    for more details on bearing capacity equation used.
    """

    @round_(ndigits=2)
    def _bearing_capacity(self) -> float:
        """Calculates ultimate bearing capacity for circular footing."""
        return (
            self._cohesion_term(1.3)
            + self._surcharge_term()
            + self._embedment_term(0.3)
        )


class TerzaghiUBC4SquareFooting(TerzaghiUltimateBearingCapacity):
    """Ultimate bearing capacity for square footing according to
    `Terzaghi 1943``.

    See [implementation](../formulas/ultimate-bearing-capacity.md/#terzaghi-bearing-capacity-for-square-footing)
    for more details on bearing capacity equation used.
    """

    def _bearing_capacity(self):
        """Calcalates ultimate bearing capacity for square footing."""
        return (
            self._cohesion_term(1.3)
            + self._surcharge_term()
            + self._embedment_term(0.4)
        )


class TerzaghiUBC4RectangularFooting(TerzaghiUltimateBearingCapacity):
    r"""Ultimate bearing capacity for rectangular footing according to
    `Terzaghi 1943`.

    See [implementation](../formulas/ultimate-bearing-capacity.md/#terzaghi-bearing-capacity-for-rectangular-footing)
    for more details on bearing capacity equation used.
    """

    @round_(ndigits=2)
    def _bearing_capacity(self) -> float:
        """Calculates ultimate bearing capacity for rectangular footing."""
        width = self.foundation_size.width
        length = self.foundation_size.length
        coh_coef = 1.0 + 0.3 * (width / length)
        emb_coef = (1.0 - 0.2 * (width / length)) / 2.0

        return (
            self._cohesion_term(coh_coef)
            + self._surcharge_term()
            + self._embedment_term(emb_coef)
        )
